@@ -14,10 +14,7 @@
 
 from datetime import datetime
 
-try:
-    import urlparse
-except ImportError:
-    import urllib.parse as urlparse
+import six.moves.urllib.parse as urlparse
 
 from cinderclient import client as base_client
 from cinderclient.tests.unit import fakes
@@ -25,22 +22,13 @@ import cinderclient.tests.unit.utils as utils
 from cinderclient.v2 import client
 
 
-def _stub_volume(**kwargs):
+REQUEST_ID = 'req-test-request-id'
+
+
+def _stub_volume(*args, **kwargs):
     volume = {
-        'id': '1234',
-        'name': None,
-        'description': None,
-        "attachments": [],
-        "bootable": "false",
-        "availability_zone": "cinder",
-        "created_at": "2012-08-27T00:00:00.000000",
-        "id": '00000000-0000-0000-0000-000000000000',
-        "metadata": {},
-        "size": 1,
-        "snapshot_id": None,
-        "status": "available",
-        "volume_type": "None",
-        "multiattach": "false",
+        "migration_status": None,
+        "attachments": [{u'server_id': u'1234'}],
         "links": [
             {
                 "href": "http://localhost/v2/fake/volumes/1234",
@@ -51,7 +39,31 @@ def _stub_volume(**kwargs):
                 "rel": "bookmark"
             }
         ],
-    }
+        "availability_zone": "cinder",
+        "os-vol-host-attr:host": "ip-192-168-0-2",
+        "encrypted": "false",
+        "updated_at": "2013-11-12T21:00:00.000000",
+        "os-volume-replication:extended_status": "None",
+        "replication_status": "disabled",
+        "snapshot_id": None,
+        'id': 1234,
+        "size": 1,
+        "user_id": "1b2d6e8928954ca4ae7c243863404bdc",
+        "os-vol-tenant-attr:tenant_id": "eb72eb33a0084acf8eb21356c2b021a7",
+        "os-vol-mig-status-attr:migstat": None,
+        "metadata": {},
+        "status": "available",
+        'description': None,
+        "multiattach": "false",
+        "os-volume-replication:driver_data": None,
+        "source_volid": None,
+        "consistencygroup_id": None,
+        "os-vol-mig-status-attr:name_id": None,
+        "name": "sample-volume",
+        "bootable": "false",
+        "created_at": "2012-08-27T00:00:00.000000",
+        "volume_type": "None",
+   }
     volume.update(kwargs)
     return volume
 
@@ -237,12 +249,72 @@ def _stub_extend(id, new_size):
     return {'volume_id': '712f4980-5ac1-41e5-9383-390aa7c9f58b'}
 
 
+def _stub_server_versions():
+    return [
+        {
+            "status": "SUPPORTED",
+            "updated": "2015-07-30T11:33:21Z",
+            "links": [
+                {
+                    "href": "http://docs.openstack.org/",
+                    "type": "text/html",
+                    "rel": "describedby",
+                },
+                {
+                    "href": "http://localhost:8776/v1/",
+                    "rel": "self",
+                }
+            ],
+            "min_version": "",
+            "version": "",
+            "id": "v1.0",
+        },
+        {
+            "status": "SUPPORTED",
+            "updated": "2015-09-30T11:33:21Z",
+            "links": [
+                {
+                    "href": "http://docs.openstack.org/",
+                    "type": "text/html",
+                    "rel": "describedby",
+                },
+                {
+                    "href": "http://localhost:8776/v2/",
+                    "rel": "self",
+                }
+            ],
+            "min_version": "",
+            "version": "",
+            "id": "v2.0",
+        },
+        {
+            "status": "CURRENT",
+            "updated": "2016-04-01T11:33:21Z",
+            "links": [
+                {
+                    "href": "http://docs.openstack.org/",
+                    "type": "text/html",
+                    "rel": "describedby",
+                },
+                {
+                    "href": "http://localhost:8776/v3/",
+                    "rel": "self",
+                }
+            ],
+            "min_version": "3.0",
+            "version": "3.1",
+            "id": "v3.0",
+        }
+    ]
+
+
 class FakeClient(fakes.FakeClient, client.Client):
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, api_version=None, *args, **kwargs):
         client.Client.__init__(self, 'username', 'password',
                                'project_id', 'auth_url',
                                extensions=kwargs.get('extensions'))
+        self.api_version = api_version
         self.client = FakeHTTPClient(**kwargs)
 
     def get_volume_api_version_from_endpoint(self):
@@ -251,7 +323,7 @@ class FakeClient(fakes.FakeClient, client.Client):
 
 class FakeHTTPClient(base_client.HTTPClient):
 
-    def __init__(self, **kwargs):
+    def __init__(self, version_header=None, **kwargs):
         self.username = 'username'
         self.password = 'password'
         self.auth_url = 'auth_url'
@@ -259,6 +331,7 @@ class FakeHTTPClient(base_client.HTTPClient):
         self.management_url = 'http://10.0.2.15:8776/v2/fake'
         self.osapi_max_limit = 1000
         self.marker = None
+        self.version_header = version_header
 
     def _cs_request(self, url, method, **kwargs):
         # Check that certain things are called correctly
@@ -293,6 +366,10 @@ class FakeHTTPClient(base_client.HTTPClient):
         # Note the call
         self.callstack.append((method, url, kwargs.get('body', None)))
         status, headers, body = getattr(self, callback)(**kwargs)
+        # add fake request-id header
+        headers['x-openstack-request-id'] = REQUEST_ID
+        if self.version_header:
+            headers['OpenStack-API-version'] = version_header
         r = utils.TestResponse({
             "status_code": status,
             "text": body,
@@ -300,11 +377,6 @@ class FakeHTTPClient(base_client.HTTPClient):
         })
         return r, body
 
-        if hasattr(status, 'items'):
-            return utils.TestResponse(status), body
-        else:
-            return utils.TestResponse({"status": status}), body
-
     def get_volume_api_version_from_endpoint(self):
         magic_tuple = urlparse.urlsplit(self.management_url)
         scheme, netloc, path, query, frag = magic_tuple
@@ -346,6 +418,10 @@ class FakeHTTPClient(base_client.HTTPClient):
             assert 'status' in body['os-reset_status']
         elif action == 'os-update_snapshot_status':
             assert 'status' in body['os-update_snapshot_status']
+        elif action == 'os-force_delete':
+            assert body[action] is None
+        elif action == 'os-unmanage':
+            assert body[action] is None
         else:
             raise AssertionError('Unexpected action: %s' % action)
         return (resp, {}, _body)
@@ -385,13 +461,9 @@ class FakeHTTPClient(base_client.HTTPClient):
                 {'id': 5678, 'name': 'sample-volume2'}
             ]})
 
-    # TODO(jdg): This will need to change
-    # at the very least it's not complete
     def get_volumes_detail(self, **kw):
         return (200, {}, {"volumes": [
-            {'id': kw.get('id', 1234),
-             'name': 'sample-volume',
-             'attachments': [{'server_id': 1234}]},
+            _stub_volume(id=kw.get('id', 1234))
         ]})
 
     def get_volumes_1234(self, **kw):
@@ -402,6 +474,10 @@ class FakeHTTPClient(base_client.HTTPClient):
         r = {'volume': self.get_volumes_detail(id=5678)[2]['volumes'][0]}
         return (200, {}, r)
 
+    def get_volumes_1234_metadata(self, **kw):
+        r = {"metadata": {'k1': 'v1', 'k2': 'v2', 'k3': 'v3'}}
+        return (200, {}, r)
+
     def get_volumes_1234_encryption(self, **kw):
         r = {'encryption_key_id': 'id'}
         return (200, {}, r)
@@ -423,7 +499,7 @@ class FakeHTTPClient(base_client.HTTPClient):
             assert body[action] is None
         elif action == 'os-initialize_connection':
             assert list(body[action]) == ['connector']
-            return (202, {}, {'connection_info': 'foos'})
+            return (202, {}, {'connection_info': {'foos': 'bars'}})
         elif action == 'os-terminate_connection':
             assert list(body[action]) == ['connector']
         elif action == 'os-begin_detaching':
@@ -431,20 +507,13 @@ class FakeHTTPClient(base_client.HTTPClient):
         elif action == 'os-roll_detaching':
             assert body[action] is None
         elif action == 'os-reset_status':
-            assert 'status' in body[action]
+            assert ('status' or 'attach_status' or 'migration_status'
+                    in body[action])
         elif action == 'os-extend':
             assert list(body[action]) == ['new_size']
         elif action == 'os-migrate_volume':
             assert 'host' in body[action]
             assert 'force_host_copy' in body[action]
-        elif action == 'os-enable_replication':
-            assert body[action] is None
-        elif action == 'os-disable_replication':
-            assert body[action] is None
-        elif action == 'os-list_replication_targets':
-            assert body[action] is None
-        elif action == 'os-failover_replication':
-            assert 'secondary' in body[action]
         elif action == 'os-update_readonly_flag':
             assert list(body[action]) == ['readonly']
         elif action == 'os-retype':
@@ -463,6 +532,9 @@ class FakeHTTPClient(base_client.HTTPClient):
             assert 'key' in body[action]
         elif action == 'os-show_image_metadata':
             assert body[action] is None
+        elif action == 'os-volume_upload_image':
+            assert 'image_name' in body[action]
+            _body = body
         else:
             raise AssertionError("Unexpected action: %s" % action)
         return (resp, {}, _body)
@@ -639,7 +711,7 @@ class FakeHTTPClient(base_client.HTTPClient):
         return (200, {}, {'volume_type': {'id': 1,
                           'name': 'test-type-1',
                           'description': 'test_type-1-desc',
-                          'extra_specs': {}}})
+                          'extra_specs': {u'key': u'value'}}})
 
     def get_types_2(self, **kw):
         return (200, {}, {'volume_type': {'id': 2,
@@ -683,12 +755,28 @@ class FakeHTTPClient(base_client.HTTPClient):
     def delete_types_1_extra_specs_k(self, **kw):
         return(204, {}, None)
 
+    def delete_types_1_extra_specs_m(self, **kw):
+        return(204, {}, None)
+
     def delete_types_1(self, **kw):
         return (202, {}, None)
 
+    def delete_types_3_extra_specs_k(self, **kw):
+        return(204, {}, None)
+
+    def delete_types_3(self, **kw):
+        return (202, {}, None)
+
     def put_types_1(self, **kw):
         return self.get_types_1()
 
+    def put_types_3(self, **kw):
+        return (200, {}, {'volume_type': {'id': 3,
+                          'name': 'test-type-2',
+                          'description': 'test_type-3-desc',
+                          'is_public': True,
+                          'extra_specs': {}}})
+
     #
     # VolumeAccess
     #
@@ -801,6 +889,12 @@ class FakeHTTPClient(base_client.HTTPClient):
     def delete_backups_76a17945_3c6f_435c_975b_b5685db10b62(self, **kw):
         return (202, {}, None)
 
+    def delete_backups_1234(self, **kw):
+        return (202, {}, None)
+
+    def delete_backups_5678(self, **kw):
+        return (202, {}, None)
+
     def post_backups(self, **kw):
         base_uri = 'http://localhost:8776'
         tenant_id = '0fa851f6668144cf9cd8c8419c1646c1'
@@ -948,6 +1042,10 @@ class FakeHTTPClient(base_client.HTTPClient):
         return (200, {},
                 {'transfer': _stub_transfer(transfer1, base_uri, tenant_id)})
 
+    def get_with_base_url(self, url, **kw):
+        server_versions = _stub_server_versions()
+        return (200, {'versions': server_versions})
+
     #
     # Services
     #
@@ -1069,11 +1167,63 @@ class FakeHTTPClient(base_client.HTTPClient):
     def put_snapshots_1234_metadata(self, **kw):
         return (200, {}, {"metadata": {"key1": "val1", "key2": "val2"}})
 
+    def get_os_volume_manage(self, **kw):
+        vol_id = "volume-ffffffff-0000-ffff-0000-ffffffffffff"
+        vols = [{"size": 4, "safe_to_manage": False, "actual_size": 4.0,
+                 "reference": {"source-name": vol_id}},
+                {"size": 5, "safe_to_manage": True, "actual_size": 4.3,
+                 "reference": {"source-name": "myvol"}}]
+        return (200, {}, {"manageable-volumes": vols})
+
+    def get_os_volume_manage_detail(self, **kw):
+        vol_id = "volume-ffffffff-0000-ffff-0000-ffffffffffff"
+        vols = [{"size": 4, "reason_not_safe": "volume in use",
+                 "safe_to_manage": False, "extra_info": "qos_setting:high",
+                 "reference": {"source-name": vol_id},
+                 "actual_size": 4.0},
+                {"size": 5, "reason_not_safe": None, "safe_to_manage": True,
+                 "extra_info": "qos_setting:low", "actual_size": 4.3,
+                 "reference": {"source-name": "myvol"}}]
+        return (200, {}, {"manageable-volumes": vols})
+
     def post_os_volume_manage(self, **kw):
         volume = _stub_volume(id='1234')
         volume.update(kw['body']['volume'])
         return (202, {}, {'volume': volume})
 
+    def get_os_snapshot_manage(self, **kw):
+        snap_id = "snapshot-ffffffff-0000-ffff-0000-ffffffffffff"
+        snaps = [{"actual_size": 4.0, "size": 4,
+                 "safe_to_manage": False, "source_id_type": "source-name",
+                 "source_cinder_id": "00000000-ffff-0000-ffff-00000000",
+                 "reference": {"source-name": snap_id},
+                 "source_identifier": "volume-00000000-ffff-0000-ffff-000000"},
+                {"actual_size": 4.3, "reference": {"source-name": "mysnap"},
+                 "source_id_type": "source-name", "source_identifier": "myvol",
+                 "safe_to_manage": True, "source_cinder_id": None, "size": 5}]
+        return (200, {}, {"manageable-snapshots": snaps})
+
+    def get_os_snapshot_manage_detail(self, **kw):
+        snap_id = "snapshot-ffffffff-0000-ffff-0000-ffffffffffff"
+        snaps = [{"actual_size": 4.0, "size": 4,
+                 "safe_to_manage": False, "source_id_type": "source-name",
+                 "source_cinder_id": "00000000-ffff-0000-ffff-00000000",
+                 "reference": {"source-name": snap_id},
+                 "source_identifier": "volume-00000000-ffff-0000-ffff-000000",
+                 "extra_info": "qos_setting:high",
+                 "reason_not_safe": "snapshot in use"},
+                {"actual_size": 4.3, "reference": {"source-name": "mysnap"},
+                 "safe_to_manage": True, "source_cinder_id": None,
+                 "source_id_type": "source-name", "identifier": "mysnap",
+                 "source_identifier": "myvol", "size": 5,
+                 "extra_info": "qos_setting:low", "reason_not_safe": None}]
+        return (200, {}, {"manageable-snapshots": snaps})
+
+    def post_os_snapshot_manage(self, **kw):
+        snapshot = _stub_snapshot(id='1234', volume_id='volume_id1')
+        snapshot.update(kw['body']['snapshot'])
+        return (202, {}, {'snapshot': snapshot})
+
     def post_os_promote_replica_1234(self, **kw):
         return (202, {}, {})
 
